/* 
 * CUAG AutoUmpire System
 * 
 * Copyright (C) 2003/2004 Matthew Johnson, Adam Biltcliffe,
 * Michael Cripps, Martin O'Leary, Edward Allcutt and James Osborn
 *
 * This program is free software; you can redistribute it and/or 
 * modify it under the terms of the GNU General Public License 
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version. 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details. 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 * To Contact the authors, please email assassins@srcf.ucam.org
 *
 */

package org.ucam.srcf.assassins.raspberry;

import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.NoSuchElementException;
import java.util.SortedMap;
import java.util.TreeMap;
import java.util.Vector;

import org.ucam.srcf.assassins.AssassinViewer;
import org.ucam.srcf.assassins.Death;
import org.ucam.srcf.assassins.Debug;
import org.ucam.srcf.assassins.Event;
import org.ucam.srcf.assassins.EventViewer;
import org.ucam.srcf.assassins.Plugin;
import org.ucam.srcf.assassins.PluginConnector;
import org.ucam.srcf.assassins.Resurrection;
import org.ucam.srcf.assassins.StateChange;
import org.ucam.srcf.assassins.Tools;
import org.ucam.srcf.assassins.exceptions.AssassinSQLException;
import org.ucam.srcf.assassins.xml.Config;

/**
 * Interface for java classes which act as plug-in game modifiers.
 */
public class Raspberry extends Plugin
{
   /** the amount to increase competence by */
   private int defaultTimeExtension;

   /* we really want to run first as we do vanialla type changes */
   public int getPriority() { return -10; }

   public static String getDescription() { return "Manages the auto-resurrection of players"; };
   
   public void setPluginConnector(PluginConnector p)
   {
      super.setPluginConnector(p);
      String temp = Config.getConfig(this.getPluginName(), "defaultTimeExtension");
      /* Config file value is in minutes, convert to milliseconds */
      defaultTimeExtension = Integer.parseInt(temp)*60*1000;
   }

   protected Vector stateChangeAdded(StateChange sc, Map state, Vector qs) throws AssassinSQLException 
   {
      if(sc instanceof Death) {
         checkResurrect(((Death) sc).getVictim());
      }
      return new Vector();
   }

   protected Vector updateState(Map state, Vector qs, EventViewer e) throws AssassinSQLException {
      checkResurrect();
      return new Vector();
   }

   /**
    * Go thorugh and ensure every death has a matching resurrection for
    * all assassins.
    * <b>But</b> if and only if they're alive at the time, we may need 
    * to remove some resurrection statechanges to do this.
    */
   private void checkResurrect() throws AssassinSQLException {
		if(Debug.debug) Debug.print(this, "checkResurrect (All)");

      SortedMap resurrectMap = new TreeMap();
      Map resurrectAt = new HashMap();
      
      /* The 1000 on the next line is just a fudge factor to ensure that
       * we get events put in to happen right this minute. */
      List statechanges = StateChange.getActiveBeforeTimeOrderedByTime(vanilla.getNowTime()+defaultTimeExtension+1000);
      /* We now have a list of all statechanges and
       * further they are returned in time order */
      
      if(Debug.debug) Debug.print(this, "Got "+statechanges.size()+" statechanges");
      Iterator i=statechanges.iterator();
      while(i.hasNext()) {
         StateChange sc = (StateChange) i.next();
         EventViewer e = sc.getEventViewer();

         /* Are we overdue a resurrection? 
          * firstKey() throws a NoSuchElementException if the map is
          * empty */
         try {
            Long next_resurrection = (Long) resurrectMap.firstKey();
            if(Debug.debug) Debug.print(this,"Checking if we are overdue for a resurrection at time "+next_resurrection);
            while(e.getTime() > next_resurrection.longValue()) {
               AssassinViewer a = (AssassinViewer) resurrectMap.get(next_resurrection);
               addEventWithResurrection(next_resurrection.longValue(), a);
               resurrectMap.remove(next_resurrection);
               resurrectAt.remove(a);
               if(Debug.debug) Debug.print(this, "We made "+a.getPlayerViewer().getName()+" rise from the dead at "+next_resurrection+" ("+Tools.prettyDateAndTime(next_resurrection.longValue())+")");
               next_resurrection = (Long) resurrectMap.firstKey();
            }
         } catch (NoSuchElementException nsee) { }

         if(Debug.debug) Debug.print(this, "Evaluating StateChange "+sc+" at time "+e.getTime()+" ("+Tools.prettyDateAndTime(e.getTime())+")");

         /* Process the statechange */
         if(sc instanceof Death) {
            AssassinViewer a = sc.getAssassinViewer();
            if(!resurrectAt.keySet().contains(a)) {
               resurrectMap.put(new Long(e.getTime()+defaultTimeExtension), a);
               resurrectAt.put(a, new Long(e.getTime()+defaultTimeExtension));
               if(Debug.debug) Debug.print(this, "Is a death of "+a.getPlayerViewer().getName()+", adding resurrection request at "+(e.getTime()+defaultTimeExtension)+" ("+Tools.prettyDateAndTime(e.getTime()+defaultTimeExtension)+")");
            }
         }
         if(sc instanceof Resurrection) {
            AssassinViewer a = sc.getAssassinViewer();
            long resurrect_at = ((Long) resurrectAt.get(a)).longValue();
            if(!resurrectAt.keySet().contains(a) || resurrect_at != e.getTime()) {
               /* Remove this resurrection */
               vanilla.removeStateChange(e, sc);
               if(Debug.debug) Debug.print(this, "Removing incorrect resurrection of "+a.getPlayerViewer().getName()+" at "+e.getTime()+" - dead till "+resurrect_at);
            } else {
               resurrectMap.remove(resurrectAt.get(a));
               resurrectAt.remove(a);
               if(Debug.debug) Debug.print(this, "Correctly resurrected "+a.getPlayerViewer().getName()+" at time "+e.getTime() + "("+Tools.prettyDateAndTime(e.getTime())+")");
            }
         }
         if(Debug.debug) Debug.print(this, "Done with StateChange "+sc);
      }
      if(Debug.debug) Debug.print(this, "Now putting in any held over resurrection events");
      
      /* Add a resurrection at some point after the last event if they're still dead */
      i = resurrectMap.keySet().iterator();
      while(i.hasNext()) {
         Long next_resurrection = (Long) i.next();
         AssassinViewer a = (AssassinViewer) resurrectMap.get(next_resurrection);
         addEventWithResurrection(next_resurrection.longValue(), a);
      }
		if(Debug.debug) Debug.print(this, "done (checkResurrect - All)");
   }

   /**
    * Go thorugh and ensure every death has a matching resurrection for the
    * given assassin.
    * <b>But</b> if and only if they're alive at the time, we may need 
    * to remove some resurrection statechanges to do this.
    */
   private void checkResurrect(AssassinViewer a) throws AssassinSQLException {
		if(Debug.debug) Debug.print(this, "checkResurrect ("+a.getPlayerViewer().getName()+")");

      boolean dead=false;
      long resurrect_at=0;
      
      /* The 1000 on the next line is just a fudge factor to ensure that
       * we get events put in to happen right this minute. */
      List statechanges = StateChange.getActiveByAssassin(a, vanilla.getNowTime()+defaultTimeExtension+1000);
      /* We now have a list of all statechanges invlolving this assassin
       * further they are returned in time order */
      
      if(Debug.debug) Debug.print(this, "Got "+statechanges.size()+" statechanges");
      Iterator i=statechanges.iterator();
      while(i.hasNext()) {
         StateChange sc = (StateChange) i.next();
         EventViewer e = sc.getEventViewer();

         /* Are we overdue a resurrection? */
         if(dead && e.getTime() > resurrect_at) {
            addEventWithResurrection(resurrect_at, a);
            dead=false;
            if(Debug.debug) Debug.print(this, "We made "+a.getPlayerViewer().getName()+" rise from the dead at "+resurrect_at+" ("+Tools.prettyDateAndTime(resurrect_at)+")");
         }

         if(Debug.debug) Debug.print(this, "Evaluating StateChange "+sc+" at time "+e.getTime()+" ("+Tools.prettyDateAndTime(e.getTime())+") while dead="+dead);

         /* Process the statechange */
         if(!dead && sc instanceof Death) {
            Death d = (Death) sc;
            /* Death involving the assassin we're iterested in as its
             * retuned by StateChange.getActiveByAssassin()
             * which is ocurring while its alive */
            dead=true;
            resurrect_at=e.getTime()+defaultTimeExtension;
            if(Debug.debug) Debug.print(this, "Is a death of our assassin, adding resurrection request at "+resurrect_at+" ("+Tools.prettyDateAndTime(resurrect_at)+")");
         }
         if(sc instanceof Resurrection) {
            /*Must be a resurrection of our assassin */
            if(!dead || resurrect_at != e.getTime()) {
               /* Remove this resurrection */
               vanilla.removeStateChange(e, sc);
               if(Debug.debug) Debug.print(this, "Removing incorrect resurrection at "+e.getTime()+" - dead till "+resurrect_at);
            } else {
               dead=false;
               if(Debug.debug) Debug.print(this, "Correctly resurrected at time "+e.getTime() + "("+Tools.prettyDateAndTime(e.getTime())+")");
            }
         }
      }
      
      /* Add a resurrection at some point after the last event if they're still dead */
      if(dead)
         addEventWithResurrection(resurrect_at, a);
		if(Debug.debug) Debug.print(this, "done (checkResurrect - "+a.getPlayerViewer().getName()+")");
   }

   private void addEventWithResurrection(long at_time, AssassinViewer who) throws AssassinSQLException {
      if(Debug.debug) Debug.print(this, "Adding resurrection of "+who.getPlayerViewer().getName()+" on a new event at time "+at_time+" ("+Tools.prettyDateAndTime(at_time)+")");
      Event n = new Event();
      n.setEventTime(at_time);
      n.setVisible(false);
      n.setActive(true);
      n.addStateChange(new Resurrection(who));
   }
}


